@@ -15,14 +15,10 @@
 #include <QtBluetooth/QLowEnergyAdvertisingParameters>
 #include <QtBluetooth/QLowEnergyConnectionParameters>
 
-
 QT_BEGIN_NAMESPACE
 
 Q_DECLARE_LOGGING_CATEGORY(QT_BT_ANDROID)
 
-// ### Needs to be moved to qjniobject.h in qtbase
-Q_DECLARE_METATYPE(QJniObject)
-
 // BT Core v5.3, 3.2.9, Vol 3, Part F
 const int BTLE_MAX_ATTRIBUTE_VALUE_SIZE = 512;
 
@@ -34,7 +30,7 @@ static QJniObject javaUuidfromQtUuid(const QBluetoothUuid& uuid)
 
     QJniObject javaString = QJniObject::fromString(output);
     QJniObject javaUuid = QJniObject::callStaticMethod<QtJniTypes::UUID>(
-                QtJniTypes::className<QtJniTypes::UUID>(), "fromString",
+                QtJniTypes::Traits<QtJniTypes::UUID>::className(), "fromString",
                 javaString.object<jstring>());
 
     return javaUuid;
@@ -97,6 +93,8 @@ void QLowEnergyControllerPrivateAndroid::init()
                 this, &QLowEnergyControllerPrivateAndroid::characteristicChanged);
         connect(hub, &LowEnergyNotificationHub::serviceError,
                 this, &QLowEnergyControllerPrivateAndroid::serviceError);
+        connect(hub, &LowEnergyNotificationHub::remoteRssiRead,
+                this, &QLowEnergyControllerPrivateAndroid::remoteRssiRead);
     }
 }
 
@@ -107,7 +105,7 @@ void QLowEnergyControllerPrivateAndroid::connectToDevice()
         return;
     }
 
-    if (!ensureAndroidPermission(BluetoothPermission::Connect)) {
+    if (!ensureAndroidPermission(QBluetoothPermission::Access)) {
         // This is unlikely to happen as a valid local adapter is a precondition
         setError(QLowEnergyController::MissingPermissionsError);
         qCWarning(QT_BT_ANDROID) << "connectToDevice() failed due to missing permissions";
@@ -246,6 +244,8 @@ void QLowEnergyControllerPrivateAndroid::writeCharacteristic(
                             "writeCharacteristic",
                             service->androidService.object<QtJniTypes::BluetoothGattService>(),
                             charUuid.object<QtJniTypes::UUID>(), payload);
+                if (result)
+                    service->characteristicList[charHandle].value = newValue;
             }
         }
     }
@@ -291,6 +291,8 @@ void QLowEnergyControllerPrivateAndroid::writeDescriptor(
                           service->androidService.object<QtJniTypes::BluetoothGattService>(),
                           charUuid.object<QtJniTypes::UUID>(), descUuid.object<QtJniTypes::UUID>(),
                           payload);
+                if (result)
+                    service->characteristicList[charHandle].descriptorList[descHandle].value = newValue;
             }
         }
     }
@@ -364,6 +366,18 @@ void QLowEnergyControllerPrivateAndroid::mtuChanged(int mtu)
     emit q->mtuChanged(mtu);
 }
 
+void QLowEnergyControllerPrivateAndroid::remoteRssiRead(int rssi, bool success)
+{
+    Q_Q(QLowEnergyController);
+    if (success) {
+        // BT Core v5.3, 7.5.4, Vol 4, Part E
+        // The LE RSSI can take values -127..127 => narrowing to qint16 is safe
+        emit q->rssiRead(rssi);
+    } else {
+        qCDebug(QT_BT_ANDROID) << "Reading remote RSSI failed";
+        setError(QLowEnergyController::RssiReadError);
+    }
+}
 
 // called if server/peripheral
 void QLowEnergyControllerPrivateAndroid::peripheralConnectionUpdated(
@@ -864,7 +878,7 @@ static QJniObject javaParcelUuidfromQtUuid(const QBluetoothUuid &uuid)
 
     QJniObject javaString = QJniObject::fromString(output);
     QJniObject parcelUuid = QJniObject::callStaticMethod<QtJniTypes::ParcelUuid>(
-                QtJniTypes::className<QtJniTypes::ParcelUuid>(), "fromString",
+                QtJniTypes::Traits<QtJniTypes::ParcelUuid>::className(), "fromString",
                 javaString.object<jstring>());
 
     return parcelUuid;
@@ -957,8 +971,7 @@ void QLowEnergyControllerPrivateAndroid::startAdvertising(const QLowEnergyAdvert
 {
     setState(QLowEnergyController::AdvertisingState);
 
-    if (!(ensureAndroidPermission(BluetoothPermission::Advertise) &&
-          ensureAndroidPermission(BluetoothPermission::Connect))) {
+    if (!ensureAndroidPermission(QBluetoothPermission::Access | QBluetoothPermission::Advertise)) {
         qCWarning(QT_BT_ANDROID) << "startAdvertising() failed due to missing permissions";
         setError(QLowEnergyController::MissingPermissionsError);
         setState(QLowEnergyController::UnconnectedState);
@@ -1026,21 +1039,21 @@ static int setupCharPermissions(const QLowEnergyCharacteristicData &charData)
             || (charData.readConstraints()
                 & QBluetooth::AttAccessConstraint::AttAuthorizationRequired)) {
             permission |= QJniObject::getStaticField<jint>(
-                                  QtJniTypes::className<QtJniTypes::BluetoothGattCharacteristic>(),
-                                  "PERMISSION_READ");
+                           QtJniTypes::Traits<QtJniTypes::BluetoothGattCharacteristic>::className(),
+                           "PERMISSION_READ");
         }
 
         if (charData.readConstraints()
             & QBluetooth::AttAccessConstraint::AttAuthenticationRequired) {
             permission |= QJniObject::getStaticField<jint>(
-                                  QtJniTypes::className<QtJniTypes::BluetoothGattCharacteristic>(),
-                                  "PERMISSION_READ_ENCRYPTED");
+                           QtJniTypes::Traits<QtJniTypes::BluetoothGattCharacteristic>::className(),
+                           "PERMISSION_READ_ENCRYPTED");
         }
 
         if (charData.readConstraints() & QBluetooth::AttAccessConstraint::AttEncryptionRequired) {
             permission |= QJniObject::getStaticField<jint>(
-                                  QtJniTypes::className<QtJniTypes::BluetoothGattCharacteristic>(),
-                                  "PERMISSION_READ_ENCRYPTED_MITM");
+                           QtJniTypes::Traits<QtJniTypes::BluetoothGattCharacteristic>::className(),
+                           "PERMISSION_READ_ENCRYPTED_MITM");
         }
     }
 
@@ -1050,33 +1063,33 @@ static int setupCharPermissions(const QLowEnergyCharacteristicData &charData)
             || (charData.writeConstraints()
                 & QBluetooth::AttAccessConstraint::AttAuthorizationRequired)) {
             permission |= QJniObject::getStaticField<jint>(
-                                  QtJniTypes::className<QtJniTypes::BluetoothGattCharacteristic>(),
-                                  "PERMISSION_WRITE");
+                           QtJniTypes::Traits<QtJniTypes::BluetoothGattCharacteristic>::className(),
+                           "PERMISSION_WRITE");
         }
 
         if (charData.writeConstraints()
             & QBluetooth::AttAccessConstraint::AttAuthenticationRequired) {
             permission |= QJniObject::getStaticField<jint>(
-                                  QtJniTypes::className<QtJniTypes::BluetoothGattCharacteristic>(),
-                                  "PERMISSION_WRITE_ENCRYPTED");
+                           QtJniTypes::Traits<QtJniTypes::BluetoothGattCharacteristic>::className(),
+                           "PERMISSION_WRITE_ENCRYPTED");
         }
 
         if (charData.writeConstraints() & QBluetooth::AttAccessConstraint::AttEncryptionRequired) {
             permission |= QJniObject::getStaticField<jint>(
-                                  QtJniTypes::className<QtJniTypes::BluetoothGattCharacteristic>(),
-                                  "PERMISSION_WRITE_ENCRYPTED_MITM");
+                           QtJniTypes::Traits<QtJniTypes::BluetoothGattCharacteristic>::className(),
+                           "PERMISSION_WRITE_ENCRYPTED_MITM");
         }
     }
 
     if (charData.properties() & QLowEnergyCharacteristic::WriteSigned) {
         if (charData.writeConstraints() & QBluetooth::AttAccessConstraint::AttEncryptionRequired) {
             permission |= QJniObject::getStaticField<jint>(
-                                  QtJniTypes::className<QtJniTypes::BluetoothGattCharacteristic>(),
-                                  "PERMISSION_WRITE_SIGNED_MITM");
+                           QtJniTypes::Traits<QtJniTypes::BluetoothGattCharacteristic>::className(),
+                           "PERMISSION_WRITE_SIGNED_MITM");
         } else {
             permission |= QJniObject::getStaticField<jint>(
-                                  QtJniTypes::className<QtJniTypes::BluetoothGattCharacteristic>(),
-                                  "PERMISSION_WRITE_SIGNED");
+                           QtJniTypes::Traits<QtJniTypes::BluetoothGattCharacteristic>::className(),
+                           "PERMISSION_WRITE_SIGNED");
         }
     }
 
@@ -1095,21 +1108,21 @@ static int setupDescPermissions(const QLowEnergyDescriptorData &descData)
             || (descData.readConstraints()
                 & QBluetooth::AttAccessConstraint::AttAuthorizationRequired)) {
             permissions |= QJniObject::getStaticField<jint>(
-                                QtJniTypes::className<QtJniTypes::BluetoothGattDescriptor>(),
-                                "PERMISSION_READ");
+                               QtJniTypes::Traits<QtJniTypes::BluetoothGattDescriptor>::className(),
+                               "PERMISSION_READ");
         }
 
         if (descData.readConstraints()
             & QBluetooth::AttAccessConstraint::AttAuthenticationRequired) {
             permissions |= QJniObject::getStaticField<jint>(
-                                QtJniTypes::className<QtJniTypes::BluetoothGattDescriptor>(),
-                                "PERMISSION_READ_ENCRYPTED");
+                               QtJniTypes::Traits<QtJniTypes::BluetoothGattDescriptor>::className(),
+                               "PERMISSION_READ_ENCRYPTED");
         }
 
         if (descData.readConstraints() & QBluetooth::AttAccessConstraint::AttEncryptionRequired) {
             permissions |= QJniObject::getStaticField<jint>(
-                                      QtJniTypes::className<QtJniTypes::BluetoothGattDescriptor>(),
-                                      "PERMISSION_READ_ENCRYPTED_MITM");
+                               QtJniTypes::Traits<QtJniTypes::BluetoothGattDescriptor>::className(),
+                               "PERMISSION_READ_ENCRYPTED_MITM");
         }
     }
 
@@ -1118,21 +1131,21 @@ static int setupDescPermissions(const QLowEnergyDescriptorData &descData)
             || (descData.readConstraints()
                 & QBluetooth::AttAccessConstraint::AttAuthorizationRequired)) {
             permissions |= QJniObject::getStaticField<jint>(
-                                QtJniTypes::className<QtJniTypes::BluetoothGattDescriptor>(),
-                                "PERMISSION_WRITE");
+                               QtJniTypes::Traits<QtJniTypes::BluetoothGattDescriptor>::className(),
+                               "PERMISSION_WRITE");
         }
 
         if (descData.readConstraints()
             & QBluetooth::AttAccessConstraint::AttAuthenticationRequired) {
             permissions |= QJniObject::getStaticField<jint>(
-                                QtJniTypes::className<QtJniTypes::BluetoothGattDescriptor>(),
-                                "PERMISSION_WRITE_ENCRYPTED");
+                               QtJniTypes::Traits<QtJniTypes::BluetoothGattDescriptor>::className(),
+                               "PERMISSION_WRITE_ENCRYPTED");
         }
 
         if (descData.readConstraints() & QBluetooth::AttAccessConstraint::AttEncryptionRequired) {
             permissions |= QJniObject::getStaticField<jint>(
-                                      QtJniTypes::className<QtJniTypes::BluetoothGattDescriptor>(),
-                                      "PERMISSION_WRITE_ENCRYPTED_MITM");
+                               QtJniTypes::Traits<QtJniTypes::BluetoothGattDescriptor>::className(),
+                               "PERMISSION_WRITE_ENCRYPTED_MITM");
         }
     }
 
@@ -1148,11 +1161,11 @@ void QLowEnergyControllerPrivateAndroid::addToGenericAttributeList(const QLowEne
 
     // create BluetoothGattService object
     jint sType = QJniObject::getStaticField<jint>(
-                            QtJniTypes::className<QtJniTypes::BluetoothGattService>(),
+                            QtJniTypes::Traits<QtJniTypes::BluetoothGattService>::className(),
                             "SERVICE_TYPE_PRIMARY");
     if (serviceData.type() == QLowEnergyServiceData::ServiceTypeSecondary)
         sType = QJniObject::getStaticField<jint>(
-                            QtJniTypes::className<QtJniTypes::BluetoothGattService>(),
+                            QtJniTypes::Traits<QtJniTypes::BluetoothGattService>::className(),
                             "SERVICE_TYPE_SECONDARY");
 
     service->androidService = QJniObject::construct<QtJniTypes::BluetoothGattService>(
@@ -1203,21 +1216,20 @@ void QLowEnergyControllerPrivateAndroid::addToGenericAttributeList(const QLowEne
         QJniEnvironment env;
         jbyteArray jb = env->NewByteArray(charData.value().size());
         env->SetByteArrayRegion(jb, 0, charData.value().size(), (jbyte*)charData.value().data());
-        jboolean success = javaChar.callMethod<jboolean>("setValue", jb);
+        jboolean success = javaChar.callMethod<jboolean>("setLocalValue", jb);
         if (!success)
             qCWarning(QT_BT_ANDROID) << "Cannot setup initial characteristic value for " << charData.uuid();
-
         env->DeleteLocalRef(jb);
 
         const QList<QLowEnergyDescriptorData> descriptorList = charData.descriptors();
         for (const auto &descData: descriptorList) {
-            QJniObject javaDesc = QJniObject::construct<QtJniTypes::BluetoothGattDescriptor>(
+            QJniObject javaDesc = QJniObject::construct<QtJniTypes::QtBtGattDescriptor>(
                                     javaUuidfromQtUuid(descData.uuid()).object<QtJniTypes::UUID>(),
                                     setupDescPermissions(descData));
 
             jb = env->NewByteArray(descData.value().size());
             env->SetByteArrayRegion(jb, 0, descData.value().size(), (jbyte*)descData.value().data());
-            success = javaDesc.callMethod<jboolean>("setValue", jb);
+            success = javaDesc.callMethod<jboolean>("setLocalValue", jb);
             if (!success) {
                 qCWarning(QT_BT_ANDROID) << "Cannot setup initial descriptor value for "
                                          << descData.uuid() << "(char" << charData.uuid()
@@ -1261,4 +1273,13 @@ int QLowEnergyControllerPrivateAndroid::mtu() const
     }
 }
 
+void QLowEnergyControllerPrivateAndroid::readRssi()
+{
+    if (!hub || !hub->javaObject().callMethod<jboolean>("readRemoteRssi")) {
+        qCWarning(QT_BT_ANDROID) << "request to read RSSI failed";
+        setError(QLowEnergyController::RssiReadError);
+        return;
+    }
+}
+
 QT_END_NAMESPACE
